from typing import Final, Literal

VerificationMode = Literal["sampled", "full"]
SolanaCluster = Literal["solana-devnet", "solana-mainnet"]

VERIFICATION_MODES: Final[tuple[VerificationMode, ...]] = ("sampled", "full")
SOLANA_CLUSTERS: Final[tuple[SolanaCluster, ...]] = ("solana-devnet", "solana-mainnet")



from __future__ import annotations

import json
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, TypedDict, cast

from trazaeo._runtime import load_trazaeo_attrs

from trazaeo_workflows.errors import ConversionRuntimeError, DependencyUnavailableError
from trazaeo_workflows.runtime_payloads import (
    load_json_object,
    require_object_field,
    require_optional_str_field,
    require_str_field,
    require_str_list_field,
)
from trazaeo_workflows.workflow_types import VerificationMode

DatasetBundleBuilder = Callable[[str], str]
DeliveryProofPackageBuilder = Callable[[str, str], str]
FullRootProofPackageBuilder = Callable[[str, int, int | None], str]
DeliveryProofReportVerifier = Callable[[str, str | None, str | None], str]


@dataclass(frozen=True)
class DatasetWorkflowBoundary:
    build_dataset_bootstrap_bundle_json: DatasetBundleBuilder | None = None
    build_dataset_incremental_bundle_json: DatasetBundleBuilder | None = None
    build_dataset_delivery_proof_package_json: DeliveryProofPackageBuilder | None = None
    full_root_proof_package_json: FullRootProofPackageBuilder | None = None
    verify_delivery_proof_package_report_json: DeliveryProofReportVerifier | None = None


class SourceFileEntry(TypedDict):
    source_uri: str
    content_hash: str
    byte_length: int
    observed_mtime: str | None


class DatasetPublishedArtifactInput(TypedDict):
    artifact_id: str
    artifact_ref: str
    content_root_hash: str
    media_type: str
    content_descriptor_ref: str | None
    content_descriptor_hash: str | None
    snapshot_id: str | None
    commit_id: str | None


class DatasetWorkflowCommonInput(TypedDict):
    schema_version: str
    issued_at: str
    signed_at: str
    signer_id: str
    signing_key_hex: str
    source_manifest_id: str
    source_dataset_id: str
    source_files: list[SourceFileEntry]
    transform_subject_id: str
    transform_job_id: str
    transform_stage: str
    publish_subject_id: str
    dataset_id: str
    dataset_version: str
    published_artifact: DatasetPublishedArtifactInput
    toolchain: str
    parameters_ref: str
    parameters_hash: str
    determinism_profile: str
    runtime_env_ref: str | None
    runtime_env_hash: str | None
    transform_spec_ref: str
    transform_spec_hash: str
    chunking_profile_ref: str | None
    chunking_profile_hash: str | None
    execution_manifest_ref: str | None
    execution_manifest_hash: str | None
    runtime_manifest_ref: str | None
    runtime_manifest_hash: str | None
    checkpoint_id: str
    checkpoint_time_window: str
    verification_policy_id: str
    stac_refs: list[str]
    reward_context_ref: str | None
    reward_context_hash: str | None
    reward_eligible: bool
    trust_policy: dict[str, Any]
    verification_mode: VerificationMode


class DatasetBootstrapWorkflowInput(TypedDict):
    common: DatasetWorkflowCommonInput
    bootstrap_origin_label: str


class DatasetProvenanceBundle(TypedDict):
    source_manifest: dict[str, Any]
    source_manifest_ref: str
    source_manifest_hash: str
    transform_envelope: dict[str, Any]
    checkpoint_manifest: dict[str, Any]
    checkpoint_manifest_ref: str
    checkpoint_manifest_hash: str
    publish_envelope: dict[str, Any]
    verification_report: dict[str, Any]
    lineage_envelopes: list[str]
    signature_bundle: list[str]
    store_ref: str
    snapshot_id: str | None
    commit_id: str | None


class DatasetIncrementalWorkflowInput(TypedDict):
    common: DatasetWorkflowCommonInput
    prior_bundle: DatasetProvenanceBundle


class DeliveryProofPackage(TypedDict):
    artifact_id: str
    checkpoint_manifest: dict[str, Any]
    lineage_envelopes: list[str]
    signature_bundle: list[str]


class FullRootProofPackage(TypedDict):
    artifact_id: str
    content_descriptor: dict[str, Any]
    content_proof_type: str
    verified_ranges: list[dict[str, Any]]


class DeliveryProofReport(TypedDict):
    package_state: str
    content_state: str
    lineage_state: str
    artifact_id: str
    artifact_path: str | None
    trusted_checkpoint_log_root_hex: str | None


def load_dataset_workflow_boundary(*attr_names: str) -> DatasetWorkflowBoundary:
    runtime = load_trazaeo_attrs(
        *attr_names,
        error_cls=DependencyUnavailableError,
    )
    return DatasetWorkflowBoundary(
        build_dataset_bootstrap_bundle_json=cast(
            DatasetBundleBuilder | None,
            runtime.get("build_dataset_bootstrap_bundle_json"),
        ),
        build_dataset_incremental_bundle_json=cast(
            DatasetBundleBuilder | None,
            runtime.get("build_dataset_incremental_bundle_json"),
        ),
        build_dataset_delivery_proof_package_json=cast(
            DeliveryProofPackageBuilder | None,
            runtime.get("build_dataset_delivery_proof_package_json"),
        ),
        full_root_proof_package_json=cast(
            FullRootProofPackageBuilder | None,
            runtime.get("full_root_proof_package_json"),
        ),
        verify_delivery_proof_package_report_json=cast(
            DeliveryProofReportVerifier | None,
            runtime.get("verify_delivery_proof_package_report_json"),
        ),
    )


def _call_runtime_json(
    operation: str,
    action: Callable[[], str],
) -> dict[str, Any]:
    try:
        payload_json = action()
    except (TypeError, ValueError, KeyError) as exc:
        raise ConversionRuntimeError(f"{operation} failed: {exc}") from exc
    return load_json_object(
        payload_json,
        operation=operation,
        error_cls=ConversionRuntimeError,
    )


def _validate_dataset_provenance_bundle(
    payload: dict[str, Any],
    *,
    operation: str,
) -> DatasetProvenanceBundle:
    return {
        "source_manifest": require_object_field(
            payload,
            key="source_manifest",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "source_manifest_ref": require_str_field(
            payload,
            key="source_manifest_ref",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "source_manifest_hash": require_str_field(
            payload,
            key="source_manifest_hash",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "transform_envelope": require_object_field(
            payload,
            key="transform_envelope",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "checkpoint_manifest": require_object_field(
            payload,
            key="checkpoint_manifest",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "checkpoint_manifest_ref": require_str_field(
            payload,
            key="checkpoint_manifest_ref",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "checkpoint_manifest_hash": require_str_field(
            payload,
            key="checkpoint_manifest_hash",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "publish_envelope": require_object_field(
            payload,
            key="publish_envelope",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "verification_report": require_object_field(
            payload,
            key="verification_report",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "lineage_envelopes": require_str_list_field(
            payload,
            key="lineage_envelopes",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "signature_bundle": require_str_list_field(
            payload,
            key="signature_bundle",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "store_ref": require_str_field(
            payload,
            key="store_ref",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "snapshot_id": require_optional_str_field(
            payload,
            key="snapshot_id",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "commit_id": require_optional_str_field(
            payload,
            key="commit_id",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
    }


def _validate_delivery_proof_package(
    payload: dict[str, Any],
    *,
    operation: str,
) -> DeliveryProofPackage:
    return {
        "artifact_id": require_str_field(
            payload,
            key="artifact_id",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "checkpoint_manifest": require_object_field(
            payload,
            key="checkpoint_manifest",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "lineage_envelopes": require_str_list_field(
            payload,
            key="lineage_envelopes",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "signature_bundle": require_str_list_field(
            payload,
            key="signature_bundle",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
    }


def _validate_full_root_proof_package(
    payload: dict[str, Any],
    *,
    operation: str,
) -> FullRootProofPackage:
    verified_ranges = payload.get("verified_ranges")
    if not isinstance(verified_ranges, list) or any(
        not isinstance(entry, dict) for entry in verified_ranges
    ):
        raise ConversionRuntimeError(
            f"{operation} omitted required list[object] field: verified_ranges"
        )
    return {
        "artifact_id": require_str_field(
            payload,
            key="artifact_id",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "content_descriptor": require_object_field(
            payload,
            key="content_descriptor",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "content_proof_type": require_str_field(
            payload,
            key="content_proof_type",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "verified_ranges": cast(list[dict[str, Any]], verified_ranges),
    }


def _validate_delivery_proof_report(
    payload: dict[str, Any],
    *,
    operation: str,
) -> DeliveryProofReport:
    return {
        "package_state": require_str_field(
            payload,
            key="package_state",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "content_state": require_str_field(
            payload,
            key="content_state",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "lineage_state": require_str_field(
            payload,
            key="lineage_state",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "artifact_id": require_str_field(
            payload,
            key="artifact_id",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "artifact_path": require_optional_str_field(
            payload,
            key="artifact_path",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
        "trusted_checkpoint_log_root_hex": require_optional_str_field(
            payload,
            key="trusted_checkpoint_log_root_hex",
            operation=operation,
            error_cls=ConversionRuntimeError,
        ),
    }


def build_dataset_bootstrap_bundle(
    input_payload: DatasetBootstrapWorkflowInput,
) -> DatasetProvenanceBundle:
    boundary = load_dataset_workflow_boundary("build_dataset_bootstrap_bundle_json")
    return _validate_dataset_provenance_bundle(
        _call_runtime_json(
            "build_dataset_bootstrap_bundle_json",
            lambda: cast(
                DatasetBundleBuilder,
                boundary.build_dataset_bootstrap_bundle_json,
            )(json.dumps(input_payload)),
        ),
        operation="build_dataset_bootstrap_bundle_json",
    )


def build_dataset_incremental_bundle(
    input_payload: DatasetIncrementalWorkflowInput,
) -> DatasetProvenanceBundle:
    boundary = load_dataset_workflow_boundary("build_dataset_incremental_bundle_json")
    return _validate_dataset_provenance_bundle(
        _call_runtime_json(
            "build_dataset_incremental_bundle_json",
            lambda: cast(
                DatasetBundleBuilder,
                boundary.build_dataset_incremental_bundle_json,
            )(json.dumps(input_payload)),
        ),
        operation="build_dataset_incremental_bundle_json",
    )


def build_dataset_delivery_proof_package(
    bundle: DatasetProvenanceBundle,
    range_proof_package: dict[str, Any],
) -> DeliveryProofPackage:
    boundary = load_dataset_workflow_boundary("build_dataset_delivery_proof_package_json")
    return _validate_delivery_proof_package(
        _call_runtime_json(
            "build_dataset_delivery_proof_package_json",
            lambda: cast(
                DeliveryProofPackageBuilder,
                boundary.build_dataset_delivery_proof_package_json,
            )(
                json.dumps(bundle),
                json.dumps(range_proof_package),
            ),
        ),
        operation="build_dataset_delivery_proof_package_json",
    )


def build_local_artifact_full_root_proof_package(
    artifact_path: str,
    *,
    chunk_size: int = 1 << 20,
    threads: int | None = None,
) -> FullRootProofPackage:
    boundary = load_dataset_workflow_boundary("full_root_proof_package_json")
    return _validate_full_root_proof_package(
        _call_runtime_json(
            "full_root_proof_package_json",
            lambda: cast(
                FullRootProofPackageBuilder,
                boundary.full_root_proof_package_json,
            )(artifact_path, chunk_size, threads),
        ),
        operation="full_root_proof_package_json",
    )


def verify_dataset_delivery_proof_report(
    delivery_proof_package: DeliveryProofPackage,
    *,
    artifact_path: str | None = None,
    trusted_checkpoint_log_root_hex: str | None = None,
) -> DeliveryProofReport:
    boundary = load_dataset_workflow_boundary("verify_delivery_proof_package_report_json")
    return _validate_delivery_proof_report(
        _call_runtime_json(
            "verify_delivery_proof_package_report_json",
            lambda: cast(
                DeliveryProofReportVerifier,
                boundary.verify_delivery_proof_package_report_json,
            )(
                json.dumps(delivery_proof_package),
                artifact_path,
                trusted_checkpoint_log_root_hex,
            ),
        ),
        operation="verify_delivery_proof_package_report_json",
    )



import json
from collections.abc import Mapping
from typing import Any, cast


def load_json_object(
    payload_json: str,
    *,
    operation: str,
    error_cls: type[Exception],
) -> dict[str, Any]:
    try:
        parsed = json.loads(payload_json)
    except json.JSONDecodeError as exc:
        raise error_cls(f"{operation} returned invalid JSON: {exc}") from exc
    if not isinstance(parsed, dict):
        raise error_cls(f"{operation} returned non-object JSON")
    return cast(dict[str, Any], parsed)


def require_str_field(
    payload: Mapping[str, object],
    *,
    key: str,
    operation: str,
    error_cls: type[Exception],
) -> str:
    value = payload.get(key)
    if not isinstance(value, str):
        raise error_cls(f"{operation} omitted required string field: {key}")
    return value


def require_optional_str_field(
    payload: Mapping[str, object],
    *,
    key: str,
    operation: str,
    error_cls: type[Exception],
) -> str | None:
    value = payload.get(key)
    if value is None:
        return None
    if not isinstance(value, str):
        raise error_cls(f"{operation} omitted required string field: {key}")
    return value


def require_object_field(
    payload: Mapping[str, object],
    *,
    key: str,
    operation: str,
    error_cls: type[Exception],
) -> dict[str, Any]:
    value = payload.get(key)
    if not isinstance(value, dict):
        raise error_cls(f"{operation} omitted required object field: {key}")
    return cast(dict[str, Any], value)


def require_str_list_field(
    payload: Mapping[str, object],
    *,
    key: str,
    operation: str,
    error_cls: type[Exception],
) -> list[str]:
    value = payload.get(key)
    if not isinstance(value, list) or any(not isinstance(item, str) for item in value):
        raise error_cls(f"{operation} omitted required list[str] field: {key}")
    return cast(list[str], value)



from __future__ import annotations

import importlib
from typing import Any

TRAZAEO_UNAVAILABLE_MESSAGE = "trazaeo Python bindings are unavailable"


def require_module(
    module_name: str,
    message: str,
    *,
    error_cls: type[Exception],
) -> Any:
    try:
        return importlib.import_module(module_name)
    except ImportError as exc:
        raise error_cls(message) from exc


def require_attr(
    module: Any,
    attr_name: str,
    message: str,
    *,
    error_cls: type[Exception],
) -> Any:
    try:
        return getattr(module, attr_name)
    except AttributeError as exc:
        raise error_cls(message) from exc


def load_module_attrs(
    module_name: str,
    missing_module_message: str,
    *,
    attrs: dict[str, str],
    error_cls: type[Exception],
) -> dict[str, Any]:
    module = require_module(
        module_name,
        missing_module_message,
        error_cls=error_cls,
    )
    return {
        attr_name: require_attr(module, attr_name, message, error_cls=error_cls)
        for attr_name, message in attrs.items()
    }


def trazaeo_attr_message(attr_name: str) -> str:
    return f"trazaeo Python bindings do not expose {attr_name}"


def load_trazaeo_attrs(
    *attr_names: str,
    error_cls: type[Exception],
) -> dict[str, Any]:
    return load_module_attrs(
        "trazaeo",
        TRAZAEO_UNAVAILABLE_MESSAGE,
        attrs={attr_name: trazaeo_attr_message(attr_name) for attr_name in attr_names},
        error_cls=error_cls,
    )



from collections.abc import Sequence

from trazaeo_workflows.workflow_cli import main as workflow_main


def main(argv: Sequence[str] | None = None) -> None:
    workflow_main(argv)


if __name__ == "__main__":
    main()



from trazaeo_workflows.dataset_provenance import (
    build_dataset_bootstrap_bundle,
    build_dataset_delivery_proof_package,
    build_dataset_incremental_bundle,
    build_local_artifact_full_root_proof_package,
    verify_dataset_delivery_proof_report,
)

__all__ = [
    "build_dataset_bootstrap_bundle",
    "build_dataset_incremental_bundle",
    "build_dataset_delivery_proof_package",
    "build_local_artifact_full_root_proof_package",
    "verify_dataset_delivery_proof_report",
]



import argparse
from collections.abc import Callable, Sequence
from typing import TypeVar

from trazaeo_workflows.workflow_types import SOLANA_CLUSTERS, VERIFICATION_MODES

CommandHandler = Callable[[argparse.Namespace], None]
ParserFactory = Callable[[], argparse.ArgumentParser]
ResultT = TypeVar("ResultT")

DEFAULT_NETCDF_CHUNK_SIZE = 4096
DEFAULT_NETCDF_THREADS = 4
DEFAULT_BLAKE3_CHUNK_SIZE = 1 << 20
DEFAULT_BLAKE3_THREADS = 4
DEMO_SIGNING_KEY_HEX = "4f3edf983ac636a65a842ce7c78d9aa706d3b113bce036f9a4f5762b76f70f18"


def _parse_args(parser: argparse.ArgumentParser, argv: Sequence[str] | None) -> argparse.Namespace:
    return parser.parse_args(list(argv) if argv is not None else None)


def exit_on_command_error(action: Callable[[], ResultT]) -> ResultT:  # noqa: UP047
    try:
        return action()
    except Exception as exc:
        from examples.python_netcdf.workflow_cli_commands import CommandFailure

        if isinstance(exc, CommandFailure):
            raise SystemExit(str(exc)) from exc
        raise


def _wrap_handler(handler: CommandHandler) -> CommandHandler:
    return lambda args: exit_on_command_error(lambda: handler(args))


def _hash_netcdf_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Hash a netCDF file")
    parser.add_argument("path", help="Path to .nc file")
    parser.add_argument("--chunk", type=int, default=DEFAULT_NETCDF_CHUNK_SIZE, help="Chunk size")
    parser.add_argument("--threads", type=int, default=DEFAULT_NETCDF_THREADS, help="Thread count")
    return parser


def _reliability_check_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Compute BLAKE3 root for reliability checks")
    parser.add_argument("path", help="Path to file")
    parser.add_argument(
        "--chunk",
        type=int,
        default=DEFAULT_BLAKE3_CHUNK_SIZE,
        help="Chunk size",
    )
    parser.add_argument("--threads", type=int, default=DEFAULT_BLAKE3_THREADS, help="Thread count")
    return parser


def _publish_demo_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Build and verify a trazaeo V1 publish envelope")
    parser.add_argument(
        "--verification-mode",
        "--mode",
        dest="verification_mode",
        default="sampled",
        choices=VERIFICATION_MODES,
    )
    parser.add_argument(
        "--trust-policy-json",
        required=True,
        help="Serialized TrustPolicy JSON used to verify the publish envelope",
    )
    return parser


def _capture_parser(*, description: str) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("--subject-id", required=True, help="Capture subject id")
    parser.add_argument(
        "--capture-actor-id",
        required=True,
        help="Attesting device or transport actor",
    )
    parser.add_argument(
        "--capture-system-id",
        required=True,
        help="Capture system, receiver pipeline, or helper-processing pipeline id",
    )
    parser.add_argument(
        "--capture-window",
        default="2026-01-01T00:00:00Z/2026-01-01T00:05:00Z",
        help="RFC3339 interval describing the capture window",
    )
    parser.add_argument("--segment-id", default="seg-1", help="Segment identifier")
    parser.add_argument(
        "--payload-text",
        default="demo-payload",
        help="UTF-8 payload text used to construct the captured segment bytes",
    )
    parser.add_argument(
        "--output-ref",
        required=True,
        help="Reference to the artifact produced by this capture event",
    )
    parser.add_argument(
        "--signer-id",
        default=None,
        help="Signer id used for the capture attestation",
    )
    parser.add_argument(
        "--signing-key-hex",
        default=DEMO_SIGNING_KEY_HEX,
        help="Hex private key for signing demo capture envelopes",
    )
    return parser


def _capture_source_parser() -> argparse.ArgumentParser:
    return _capture_parser(description="Build a demo source-device capture envelope")


def _capture_transport_parser() -> argparse.ArgumentParser:
    parser = _capture_parser(description="Build a demo transport-receipt capture envelope")
    parser.add_argument(
        "--input-ref",
        action="append",
        default=None,
        help="Reference to the upstream feed, pass, or prior transport handoff",
    )
    return parser


def _publish_solana_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=(
            "Build, commit, and verify a trazaeo V1 publish envelope "
            "with a Solana proof-log provider"
        )
    )
    parser.add_argument(
        "--verification-mode",
        "--mode",
        dest="verification_mode",
        default="sampled",
        choices=VERIFICATION_MODES,
    )
    parser.add_argument("--cluster", default="solana-devnet", choices=SOLANA_CLUSTERS)
    parser.add_argument("--require-proof-log", action="store_true")
    parser.add_argument(
        "--trust-policy-json",
        required=True,
        help="Serialized TrustPolicy JSON used to verify the publish envelope",
    )
    parser.add_argument(
        "--rpc-url",
        default=None,
        help="Override the Solana RPC URL",
    )
    parser.add_argument(
        "--proof-log-keypair-path",
        default=None,
        help="Path to a funded Solana keypair JSON file for the proof-log attestor",
    )
    return parser


def _icechunk_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Convert an NC collection to Zarr/Icechunk provenance artifacts"
    )
    parser.add_argument("sources", nargs="+", help="Input netCDF file paths")
    parser.add_argument("--zarr-store", required=True, help="Output Zarr store path")
    parser.add_argument("--dataset-id", required=True, help="Dataset ID")
    parser.add_argument("--dataset-version", required=True, help="Dataset version")
    parser.add_argument("--signer-id", default="publisher", help="Attestation signer id")
    parser.add_argument(
        "--signing-key-hex",
        default=DEMO_SIGNING_KEY_HEX,
        help="Hex private key for signing test/demo envelopes",
    )
    parser.add_argument(
        "--trust-policy-json",
        required=True,
        help="Serialized TrustPolicy JSON used to verify the publish envelope",
    )
    parser.add_argument("--verification-mode", choices=VERIFICATION_MODES, default="full")
    return parser


def _icechunk_lineage_demo_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=(
            "Build bootstrap and incremental Icechunk provenance bundles and verify "
            "a delivery proof for a local artifact"
        )
    )
    parser.add_argument("--bootstrap-sources", nargs="+", required=True)
    parser.add_argument("--incremental-sources", nargs="+", required=True)
    parser.add_argument("--artifact-path", required=True)
    parser.add_argument("--bootstrap-store", required=True, help="Bootstrap Icechunk store path")
    parser.add_argument(
        "--incremental-store",
        required=True,
        help="Incremental Icechunk store path",
    )
    parser.add_argument("--dataset-id", required=True, help="Dataset ID")
    parser.add_argument("--bootstrap-version", default="v1", help="Bootstrap dataset version")
    parser.add_argument(
        "--incremental-version",
        default="v2",
        help="Incremental dataset version",
    )
    parser.add_argument("--signer-id", default="publisher", help="Attestation signer id")
    parser.add_argument(
        "--signing-key-hex",
        default=DEMO_SIGNING_KEY_HEX,
        help="Hex private key for signing test/demo envelopes",
    )
    parser.add_argument(
        "--trust-policy-json",
        required=True,
        help="Serialized TrustPolicy JSON used to verify publish and delivery proofs",
    )
    parser.add_argument("--verification-mode", choices=VERIFICATION_MODES, default="full")
    return parser


def _run_hash_netcdf_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.netcdf_hashing import hash_netcdf_content_root
    from examples.python_netcdf.workflow_cli_commands import run_hash_netcdf

    run_hash_netcdf(args, hash_runner=hash_netcdf_content_root)


def _run_reliability_check_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.file_hashing import hash_file_root
    from examples.python_netcdf.workflow_cli_commands import run_reliability_check

    run_reliability_check(args, hash_runner=hash_file_root)


def _run_publish_demo_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.workflow_cli_commands import run_publish_demo

    run_publish_demo(args)


def _run_capture_source_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.workflow_cli_commands import run_capture_source

    run_capture_source(args)


def _run_capture_transport_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.workflow_cli_commands import run_capture_transport

    run_capture_transport(args)


def _run_publish_solana_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.workflow_cli_commands import run_publish_solana

    run_publish_solana(args)


def _run_icechunk_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.icechunk_conversion import convert_nc_collection_to_zarr_icechunk
    from examples.python_netcdf.workflow_cli_commands import run_icechunk

    run_icechunk(args, converter=convert_nc_collection_to_zarr_icechunk)


def _run_icechunk_lineage_demo_handler(args: argparse.Namespace) -> None:
    from examples.python_netcdf.workflow_cli_commands import run_icechunk_lineage_demo

    run_icechunk_lineage_demo(args)


def _run_single_command(
    parser_factory: ParserFactory,
    handler: CommandHandler,
    argv: Sequence[str] | None,
) -> None:
    args = _parse_args(parser_factory(), argv)
    exit_on_command_error(lambda: handler(args))


def run_hash_netcdf_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_hash_netcdf_parser, _run_hash_netcdf_handler, argv)


def run_reliability_check_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_reliability_check_parser, _run_reliability_check_handler, argv)


def run_publish_demo_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_publish_demo_parser, _run_publish_demo_handler, argv)


def run_capture_source_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_capture_source_parser, _run_capture_source_handler, argv)


def run_capture_transport_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_capture_transport_parser, _run_capture_transport_handler, argv)


def run_publish_solana_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_publish_solana_parser, _run_publish_solana_handler, argv)


def run_icechunk_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_icechunk_parser, _run_icechunk_handler, argv)


def run_icechunk_lineage_demo_cli(argv: Sequence[str] | None = None) -> None:
    _run_single_command(_icechunk_lineage_demo_parser, _run_icechunk_lineage_demo_handler, argv)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="trazaeo workflow CLI")
    subparsers = parser.add_subparsers(dest="command", required=True)

    hash_parser = subparsers.add_parser(
        "hash-netcdf",
        parents=[_hash_netcdf_parser()],
        add_help=False,
    )
    hash_parser.set_defaults(handler=_wrap_handler(_run_hash_netcdf_handler))

    reliability_parser = subparsers.add_parser(
        "reliability-check",
        parents=[_reliability_check_parser()],
        add_help=False,
    )
    reliability_parser.set_defaults(handler=_wrap_handler(_run_reliability_check_handler))

    publish_demo_parser = subparsers.add_parser(
        "publish-demo",
        parents=[_publish_demo_parser()],
        add_help=False,
    )
    publish_demo_parser.set_defaults(handler=_wrap_handler(_run_publish_demo_handler))

    capture_source_parser = subparsers.add_parser(
        "capture-source",
        parents=[_capture_source_parser()],
        add_help=False,
    )
    capture_source_parser.set_defaults(handler=_wrap_handler(_run_capture_source_handler))

    capture_transport_parser = subparsers.add_parser(
        "capture-transport",
        parents=[_capture_transport_parser()],
        add_help=False,
    )
    capture_transport_parser.set_defaults(handler=_wrap_handler(_run_capture_transport_handler))

    publish_solana_parser = subparsers.add_parser(
        "publish-solana",
        parents=[_publish_solana_parser()],
        add_help=False,
    )
    publish_solana_parser.set_defaults(handler=_wrap_handler(_run_publish_solana_handler))

    icechunk_parser = subparsers.add_parser(
        "icechunk",
        parents=[_icechunk_parser()],
        add_help=False,
    )
    icechunk_parser.set_defaults(handler=_wrap_handler(_run_icechunk_handler))

    lineage_demo_parser = subparsers.add_parser(
        "icechunk-lineage-demo",
        parents=[_icechunk_lineage_demo_parser()],
        add_help=False,
    )
    lineage_demo_parser.set_defaults(handler=_wrap_handler(_run_icechunk_lineage_demo_handler))
    return parser


def main(argv: Sequence[str] | None = None) -> None:
    parser = build_parser()
    args = _parse_args(parser, argv)
    args.handler(args)
